unit ulazQuadTreeTestMain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Windows, Graphics, Dialogs, StdCtrls, ExtCtrls,
  ComCtrls, mvMapViewer,  Contnrs, IniFiles,
  ulazQuadTree, uOSMQuadTree, uAreaHitThread,
  uWnPnPoly,
  mvGeoMath, mvGeoNames, mvTypes, mvGpsObj, mvDrawingEngine,
  mvEngine,
  math
  ;
const
  MaxItemsPerTileSqr = 4; // should be 2,4,8, other values will be low performant
  MaxItemsPerTile = MaxItemsPerTileSqr*MaxItemsPerTileSqr;

type
  TMouseRectHit = (mrhNone,
                 mrhTestArea,
                 mrhTestTop, mrhTestBottom, mrhTestRight, mrhTestLeft,
                 mrhTestNW, mrhTestNE, mrhTestSE, mrhTestSW,
                 mrhBaseArea,
                 mrhBaseTop, mrhBaseBottom, mrhBaseRight, mrhBaseLeft,
                 mrhBaseNW, mrhBaseNE, mrhBaseSE, mrhBaseSW
                 );


  { TTestPOI }

  TTestPOI = class(TObject)
  private
    FLat : Double;
    FLon : Double;
    FCaption : String;
  public
    property Lat : Double read FLat;
    property Lon : Double read FLon;
    property Caption : String read FCaption;
    constructor Create(const ALat, ALon : Double; const ACaption : String);
  end;

  { TTestShape }

  TTestShape = class(TTestPOI)
  private
    FShape : array of Double;
    FShapeBounds : TLazQuadNodeArea;
    function GetCoordsCount : Integer;
    function GetCoordsLats(CoordIndex : Integer) : Double;
    function GetCoordsLons(CoordIndex : Integer) : Double;
    function GetShapeBounds : TLazQuadNodeArea;
  public
    property CoordsCount : Integer read GetCoordsCount;
    property CoordsLats[CoordIndex : Integer] : Double read GetCoordsLats;
    property CoordsLons[CoordIndex : Integer] : Double read GetCoordsLons;
    property ShapeBounds : TLazQuadNodeArea read GetShapeBounds;
    function ReadShapesFromString(const AShapeString : String) : Boolean;
  end;

  { TMyPoint }

  TMyPoint = class(TObject)
  private
    FX : Double;
    FY : Double;
  public
    property X : Double read FX write FX;
    property Y : Double read FY write FY;
    procedure PaintToCanvas(const ACanvas : TCanvas);
    constructor Create(const AX, AY : Double);
  end;
  { TGpsPointGroupMarker
    Is displayed as a circle with the amount of grouped items
  }
  TGpsPointGroupMarker = class(TGPSPoint)
  private
    FGroupedItemsCount : Integer;
  public
    property GroupedItemsCount : Integer read FGroupedItemsCount write FGroupedItemsCount;
  end;

  { TForm1 }

  TForm1 = class(TForm)
    btnAreaHitTest: TButton;
    btnSampleImport: TButton;
    btnPolyHitTest: TButton;
    btnTreeAsStrings: TButton;
    btnRectRelief: TButton;
    Button1: TButton;
    cbPOISample: TCheckBox;
    cbShapeSample: TCheckBox;
    cbPOIInsertLazMap: TCheckBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    lblListItemsCount: TLabel;
    lblPOIDesc: TLabel;
    lblSampleImportResult: TLabel;
    lblHitTestCounter: TLabel;
    MapView: TMapView;
    Memo1: TMemo;
    Memo2: TMemo;
    Memo3: TMemo;
    Memo4: TMemo;
    PageControl1: TPageControl;
    Panel1: TPanel;
    Panel2: TPanel;
    pbEarth: TPaintBox;
    sbMapStatus: TStatusBar;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    TabSheet3: TTabSheet;
    TabSheet4: TTabSheet;
    tsSampleImport: TTabSheet;
    tsThreadedAreaHitTest: TTabSheet;
    Timer1: TTimer;
    procedure btnAreaHitTestClick(Sender: TObject);
    procedure btnRectReliefClick(Sender: TObject);
    procedure btnSampleImportClick(Sender: TObject);
    procedure btnPolyHitTestClick(Sender: TObject);
    procedure btnTreeAsStringsClick(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure MapViewChange(Sender: TObject);
    procedure MapViewDrawGpsPoint(Sender: TObject;
      ADrawer: TMvCustomDrawingEngine; APoint: TGpsPoint);
    procedure MapViewMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MapViewMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure MapViewMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbEarthMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbEarthMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure pbEarthMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbEarthPaint(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    FLatLonHitThread : TLatLonHitThread;
    FWorldmap : TBitmap;
    FTestQuadTree : TOSMQuadTree;
    FMouseIsDown : Boolean;
    FMouseDownHit : TMouseRectHit;
    FMouseDownLastX, FMouseDownLastY : Integer;
    FMouseDownLatLonRect : TLazQuadNodeArea;
    FBaseRect : TLazQuadNodeArea;
    FTestRect : TLazQuadNodeArea;
    FRectGUIUpdateInProgress : Boolean;

    FCountryShapes : TObjectList;
    FLastMapArea : TRealArea;
    FLastMapCenter : TRealPoint;
    FMapViewMouseDownX : Integer;
    FMapViewMouseDownY : Integer;


    function OnTestQuadTreeGetItemCoord(Sender : TObject; AItem : TObject; var AX, AY : Double) : Boolean;
    function OnTestQuadTreeGetItemAreaEvent(Sender : TObject; AItem : TObject; var AItemArea : TLazQuadNodeArea) : Boolean;
    function OnTestQuadTreeGetItemCaptionEvent(Sender : TObject; AItem : TObject; var AItemCaption : String) : Boolean;
    function LatLonToCavasPoint(const ALat, ALon : Double) : TPoint;
    function LatLonRectToCanvasRect(const ALatLonRect : TLazQuadNodeArea;
               var ARect, AWrapRect : TRect) : Boolean;
    function PaintMouseHitTest(const X,Y : Integer; const AShift: TShiftState) :   TMouseRectHit;
    procedure LatLonRectToGUI;
    procedure IncLatLonRect(var ALatLonRect : TLazQuadNodeArea; AInc : Double);
    procedure OnMapViewBeforeDrawObjects(Sender : TObject);
    procedure CalculateMapPixelCoords(var ALeft,ATop,ARight,ABottom : Int64; var AOSMMapSize : Int64; var AOfsX, AOfsY : Integer);
  public
    procedure ColorEarthMap(const ACanvas : TCanvas);
  end;

const
  TheWorld : TLazQuadNodeArea =
    (Left:  -180.0;
     Top:     90.0;
     Right:  180.0;
     Bottom: -90.0);

var
  Form1: TForm1;

implementation

{$R *.lfm}

const
  HitString : array[TLazQuadNodeAreaOverlap] of String = (
    'None', 'Partly', 'Inside', 'Equal', 'Full');

  ColorSet : array[0..15] of TColor = (
{
*** Primary color:

   shade 0 = #62CCF5 = rgb( 98,204,245) = rgba( 98,204,245,1) = rgb0(0.384,0.8,0.961)
   shade 1 = #E0F5FC = rgb(224,245,252) = rgba(224,245,252,1) = rgb0(0.878,0.961,0.988)
   shade 2 = #96DCF6 = rgb(150,220,246) = rgba(150,220,246,1) = rgb0(0.588,0.863,0.965)
   shade 3 = #32C0F7 = rgb( 50,192,247) = rgba( 50,192,247,1) = rgb0(0.196,0.753,0.969)
   shade 4 = #07B7FA = rgb(  7,183,250) = rgba(  7,183,250,1) = rgb0(0.027,0.718,0.98)
}
$F5CC62,$FCF5E0,$F6DC96,$F7C032,

{
  *** Secondary color (1):

     shade 0 = #FD5E86 = rgb(253, 94,134) = rgba(253, 94,134,1) = rgb0(0.992,0.369,0.525)
     shade 1 = #FEE1E8 = rgb(254,225,232) = rgba(254,225,232,1) = rgb0(0.996,0.882,0.91)
     shade 2 = #FD97B0 = rgb(253,151,176) = rgba(253,151,176,1) = rgb0(0.992,0.592,0.69)
     shade 3 = #FD2B5F = rgb(253, 43, 95) = rgba(253, 43, 95,1) = rgb0(0.992,0.169,0.373)
     shade 4 = #FE0040 = rgb(254,  0, 64) = rgba(254,  0, 64,1) = rgb0(0.996,0,0.251)
}
$865EFD,$E8E1FE,$B097FD,$5F2BFD,

{
  *** Secondary color (2):

     shade 0 = #9CFC5E = rgb(156,252, 94) = rgba(156,252, 94,1) = rgb0(0.612,0.988,0.369)
     shade 1 = #ECFEE1 = rgb(236,254,225) = rgba(236,254,225,1) = rgb0(0.925,0.996,0.882)
     shade 2 = #BEFC96 = rgb(190,252,150) = rgba(190,252,150,1) = rgb0(0.745,0.988,0.588)
     shade 3 = #7DFC2A = rgb(125,252, 42) = rgba(125,252, 42,1) = rgb0(0.49,0.988,0.165)
     shade 4 = #64FD00 = rgb(100,253,  0) = rgba(100,253,  0,1) = rgb0(0.392,0.992,0)
}

  $5EFC9C,$E1FEEC,$96FCBE,$2AFC7D,

{
  *** Complement color:

     shade 0 = #FFB25F = rgb(255,178, 95) = rgba(255,178, 95,1) = rgb0(1,0.698,0.373)
     shade 1 = #FFF1E2 = rgb(255,241,226) = rgba(255,241,226,1) = rgb0(1,0.945,0.886)
     shade 2 = #FFCE98 = rgb(255,206,152) = rgba(255,206,152,1) = rgb0(1,0.808,0.596)
     shade 3 = #FF992B = rgb(255,153, 43) = rgba(255,153, 43,1) = rgb0(1,0.6,0.169)
     shade 4 = #FF8500 = rgb(255,133,  0) = rgba(255,133,  0,1) = rgb0(1,0.522,0)
}
  $5FB2FF,$E2F1FF,$98CEFF,$2B99FF);

function ValidateCoordLine(const ACoordLine: String): Boolean;
var
  sl : TStringList;
  i : Integer;
  s : String;
  lats,lons : String;
  lat, lon : Double;
  posi : Integer;
  fs : TFormatSettings;
begin
  Result := False;
  fs.ThousandSeparator:=#0;
  fs.DecimalSeparator:= '.';
  sl := TStringList.Create;
  try
    sl.Delimiter:=';';
    sl.StrictDelimiter := True;
    sl.QuoteChar := #0;
    sl.DelimitedText := ACoordLine;
    if sl.Count < 2 then Exit; // Needs a least two pairs of Coordinates
    for i := 0 to sl.Count-1 do
    begin
      s := sl[i];
      posi := Pos(',',s);
      if posi < 2 then Exit; //at least one digit before a comma is expected
      lons := Copy(s,1,posi-1);
      lats := Copy(s,posi+1,MaxInt);
      lat := StrToFloatDef(lats,-1000,fs);
      lon := StrToFloatDef(lons,-1000,fs);
      if (lat < -90) or (lat > 90) or
         (lon < -180) or (lon > 180) then Exit;
    end;
    Result := True;
  finally
    sl.Free;
  end;
end;


{ TTestShape }

function TTestShape.GetCoordsCount: Integer;
begin
  Result := Length(FShape) div 2;
end;

function TTestShape.GetCoordsLats(CoordIndex: Integer): Double;
begin
  Result := FShape[(CoordIndex shl 1)+1];
end;

function TTestShape.GetCoordsLons(CoordIndex: Integer): Double;
begin
  Result := FShape[(CoordIndex shl 1)];
end;

function TTestShape.GetShapeBounds: TLazQuadNodeArea;
begin
  Result := FShapeBounds;
end;

function TTestShape.ReadShapesFromString(const AShapeString: String): Boolean;
var
  sl1 : TStringList;
  j : Integer;
  s : String;
  posi : Integer;
  fs : TFormatSettings;
  lats,lons : String;
  llat, llon : Double;
begin
  Result := False;
  fs.ThousandsSeparator := #0;
  fs.DecimalSeparator:= '.';
  sl1 := TStringList.Create;
  try
    sl1.Delimiter:=';';
    sl1.StrictDelimiter := True;
    sl1.QuoteChar := #0;
    sl1.DelimitedText := AShapeString;
    SetLength(FShape,sl1.Count*2);
    for j := 0 to sl1.Count-1 do
    begin
      s := sl1[j];
      posi := Pos(',',s);
      lons := Copy(s,1,posi-1);
      lats := Copy(s,posi+1,MaxInt);
      llat := StrToFloat(lats,fs);
      llon := StrToFloat(lons,fs);
      FShape[j*2] := llon;
      FShape[j*2+1] := llat;
      if j = 0 then
      begin
        FShapeBounds.Left:=llon;
        FShapeBounds.Right:=llon;
        FShapeBounds.Top:=llat;
        FShapeBounds.Bottom:=llat;
      end
      else
      begin
        if FShapeBounds.Left > llon then
          FShapeBounds.Left:=llon;
        if FShapeBounds.Right < llon then
          FShapeBounds.Right:=llon;
        if FShapeBounds.Top < llat then
          FShapeBounds.Top:=llat;
        if FShapeBounds.Bottom > llat then
          FShapeBounds.Bottom:=llat;
      end;
      Result := True;
      FLat := (FShapeBounds.Top + FShapeBounds.Bottom) / 2.0;
      FLon := (FShapeBounds.Left + FShapeBounds.Right) / 2.0;

    end;
  finally
    if Assigned(sl1) then
      sl1.Free;
  end;
end;



{ TTestPOI }

constructor TTestPOI.Create(const ALat, ALon: Double; const ACaption: String);
begin
  inherited Create;
  FLat := ALat;
  FLon := ALon;
  FCaption := ACaption;
end;

{ TMyPoint }

procedure TMyPoint.PaintToCanvas(const ACanvas: TCanvas);
var
  lx, ly : Integer;
begin
  lx := Round(FX);
  ly := Round(FY);
  ACanvas.Pen.Color := clRed;
  ACanvas.Pen.Style := psSolid;
  ACanvas.MoveTo(lx-10,ly);
  ACanvas.LineTo(lx+10,ly);
  ACanvas.MoveTo(lx,ly-10);
  ACanvas.LineTo(lx,ly+10);
end;

constructor TMyPoint.Create(const AX, AY: Double);
begin
  inherited Create;
  FX := AX;
  FY := AY;
end;


{ TForm1 }

procedure TForm1.IncLatLonRect(var ALatLonRect : TLazQuadNodeArea; AInc : Double);
begin
  NormalizeAndUnfoldArea(ALatLonRect,qydUpPlus,True,FTestQuadTree.RootNode.NodeArea);
  ALatLonRect.Left  := ALatLonRect.Left-AInc;
  ALatLonRect.Right  := ALatLonRect.Right+AInc;
  ALatLonRect.Top := ALatLonRect.Top-AInc;
  ALatLonRect.Bottom := ALatLonRect.Bottom+AInc;
  NormalizeAndUnfoldArea(ALatLonRect,qydUpPlus,True,FTestQuadTree.RootNode.NodeArea);
end;

procedure TForm1.OnMapViewBeforeDrawObjects(Sender: TObject);
var
  gpsPt : TGpsPoint;
  lItemGroupMarker : TGpsPointGroupMarker;
  maparea : TRealArea;
  llon, llat : Double;
  x, y : Integer;
  ofsx, ofsy : Integer;
  pt0, pt1, pt2 : TPoint;
  ptofs : TPoint;
  i, j : Integer;
  rpt0, rpt1, rpt2 : TRealPoint;
  n0, n1 : TLazQuadNode;
  lNodeArea : TLazQuadNodeArea;
  itemcnt : Integer;
  estimateditemcnt : Double;
  wlon0, wlon1, wlon, wlat : Double;
  wlonpx, wlatpx : Int64;
  lList : TList;
  lpoi : TTestPOI;
  tilecntx, tilecnty : Integer;
  s : String;
  pixelsperitem : Integer;
  mapwidthpx : Int64;
  lLeft,lTop,lRight,lBottom : Int64;
begin
{
  if MapView.Engine.CrossesDateline then
    s := 'True'
  else
    s := 'False';
  maparea := MapView.GetVisibleArea;
  Label4.Caption := Format('LeftLon: %1.1f° RightLon: %1.1f°, CrossesDateLine %s, OfsX %d, OfsY %d',
                             [maparea.TopLeft.Lon,maparea.BottomRight.Lon,s, ofsx,ofsy]);
}


  // Check if the Map has changed or moved in the mean time.
  // If not, exit.
  maparea := MapView.GetVisibleArea;
  rpt0 := MapView.Center;
  if FLastMapArea.Equal(maparea) and
    (FLastMapCenter.Lat = rpt0.Lat) and
    (FLastMapCenter.Lon = rpt0.Lon)  then Exit;

  FLastMapCenter.InitXY(rpt0.Lon,rpt0.Lat);
  FLastMapArea.Init(maparea.TopLeft,maparea.BottomRight);

  CalculateMapPixelCoords(lLeft,lTop,lRight,lBottom,mapwidthpx,ofsx,ofsy);
  tilecntx := (lRight-lLeft) div TILE_SIZE;
  tilecnty := (lBottom-lTop) div TILE_SIZE;

  MapView.GPSItems.Clear(0);

  pixelsperitem := OSMTileSize div MaxItemsPerTileSqr;
  for y := 1 to tilecnty*MaxItemsPerTileSqr do
  begin
    for x := 1 to tilecntx*MaxItemsPerTileSqr do
    begin
      pt2.X := lLeft+((x-1)*pixelsperitem);
      if pt2.X < 0 then
        pt2.X := pt2.X + mapwidthpx
      else if pt2.X > mapwidthpx then
        pt2.X := pt2.X - mapwidthpx;
      pt2.Y := lTop+((y-1)*pixelsperitem);
      PixelsToDegreesOSM (pt2.X,pt2.Y,MapView.Zoom,rpt2.Lat,rpt2.Lon);
      pt1.X := lLeft+(x*pixelsperitem);
      if pt1.X < 0 then
        pt1.X := pt1.X + mapwidthpx
      else if pt1.X > mapwidthpx then
        pt1.X := pt1.X - mapwidthpx;
      pt1.Y := lTop+(y*pixelsperitem);
      PixelsToDegreesOSM (pt1.X,pt1.Y,MapView.Zoom,rpt1.Lat,rpt1.Lon);
{
      gpsPt := TGpsPoint.Create(rpt2.Lon,rpt2.Lat);
      gpsPt.Name := Format('X %d, Y %d',[x,y]) ;
      MapView.GPSItems.Add(gpspt,0);
}

      lNodeArea.Bottom:=rpt1.Lat;
      lNodeArea.Right:=rpt1.Lon;
      lNodeArea.Top:=rpt2.Lat;
      lNodeArea.Left:=rpt2.Lon;

      estimateditemcnt := FTestQuadTree.EstimatedCountOfAssignedItemsInArea(lNodeArea);
      if (estimateditemcnt > 4) and (MapView.Zoom < MapView.ZoomMax) then
      begin
        lItemGroupMarker := TGpsPointGroupMarker.Create((rpt1.Lon+rpt2.Lon)/2,(rpt1.Lat+rpt2.Lat)/2.0);
        lItemGroupMarker.GroupedItemsCount := Round(estimateditemcnt);
        MapView.GPSItems.Add(lItemGroupMarker,0);
      end
      else
      begin
        itemcnt := FTestQuadTree.CountAssignedItemsInArea(lNodeArea);
        if itemcnt > 0 then
        begin
          lList := TList.Create;
          try
            FTestQuadTree.ListItemsInArea(lNodeArea,lList);
            for i := 0 to lList.Count-1 do
            begin
              lpoi := TTestPOI(lList.Items[i]);
              if Assigned(lpoi) then
              begin
                gpsPt := TGpsPoint.Create(lpoi.Lon,lpoi.Lat);
                gpsPt.Name := lpoi.Caption;
                MapView.GPSItems.Add(gpspt,0);
              end;
            end;
          finally
            lList.Free;
          end;
        end;
      end;
    end;
  end;
  lblListItemsCount.Caption := IntToStr(MapView.GPSItems.Count);


(*
  // Calculating the width in Degree of the visible Map
  // In the cyclic map several cases has to be catched
  if (maparea.BottomRight.Lon >= 180) and
     (maparea.TopLeft.Lon <= -180) then
    wlon := 360.0
  else
  begin
    if MapView.Engine.CrossesDateLine then
      maparea.TopLeft.Lon := maparea.TopLeft.Lon -360.0;
    if maparea.BottomRight.Lon < maparea.TopLeft.Lon then
      wlon := 180.0+maparea.BottomRight.Lon-maparea.TopLeft.Lon
    else
      wlon := maparea.BottomRight.Lon-maparea.TopLeft.Lon;
  end;
  wlat := maparea.TopLeft.Lat-maparea.BottomRight.Lat;
  // Calculate the offset of the tile left of the map

  pt0 := DegreesToPixelsOSM(MapView.Zoom,wlat / 2.0,wlon-180.0);
  wlonpx := pt0.X;
  mapwidthpx := round(mvGeoMath.ZoomFactor(MapView.Zoom)) * TILE_SIZE;
  if wlonpx > mapwidthpx then
    wlonpx := mapwidthpx;
  wlatpx := pt0.Y * 2;
  pt0 := DegreesToPixelsOSM(MapView.Zoom,OSMMaxLatitude,-180.0);
  wlonpx := wlonpx-pt0.X;
  wlatpx := wlatpx-pt0.Y;



{
  ptofs := DegreesToPixelsOSM(MapView.Zoom,maparea.TopLeft.Lat,maparea.TopLeft.Lon);
  ofsx := -(ptofs.X mod TILE_SIZE);
  ofsy := -(ptofs.Y mod TILE_SIZE);
}
  ptofs := DegreesToPixelsOSM(MapView.Zoom,MapView.Center.Lat,MapView.Center.Lon);
  ptofs.X := ptofs.X-(MapView.Width div 2);
  ptofs.Y := ptofs.Y-(MapView.Height div 2);
  ofsx := -(ptofs.X mod TILE_SIZE);
  ofsy := -(ptofs.Y mod TILE_SIZE);
  if ofsx > 0 then
    ofsx := ofsx - TILE_SIZE;
  if ofsy > 0 then
    ofsy := ofsy - TILE_SIZE;

  tilecntx := (wlonpx div 256) + 3;
  if wlat >= 180 then
    tilecnty := (wlatpx div 256) + 2
  else
    tilecnty := (MapView.Height div 256)+2;

  pt0.X := ofsx;
  pt0.Y := ofsy;
  rpt0 := MapView.Engine.ScreenToLatLon(pt0);
  rpt2 := rpt0;


  MapView.GPSItems.Clear(0);

  pixelsperitem := OSMTileSize div MaxItemsPerTileSqr;
  for y := 1 to tilecnty*MaxItemsPerTileSqr do
  begin
    for x := 1 to tilecntx*MaxItemsPerTileSqr do
    begin
      pt2.X := ((x-1)*pixelsperitem)+ofsx;
      if (pt2.X > mapwidthpx) or
         (pt2.X > MapView.Width) then Break;
      pt2.Y := ((y-1)*pixelsperitem)+ofsy;
      rpt2 := MapView.Engine.ScreenToLatLon(pt2);
      pt1.X := (x*pixelsperitem)+ofsx;
      if (pt1.X > mapwidthpx) or
         (pt1.X > MapView.Width) then Break;
      pt1.Y := (y*pixelsperitem)+ofsy;
      rpt1 := MapView.Engine.ScreenToLatLon(pt1);
      if rpt1.Lon < rpt2.Lon then
        rpt1.Lon := rpt1.Lon + 360.0;
      lNodeArea.Bottom:=rpt1.Lat;
      lNodeArea.Right:=rpt1.Lon;
      lNodeArea.Top:=rpt2.Lat;
      lNodeArea.Left:=rpt2.Lon;

      estimateditemcnt := FTestQuadTree.EstimatedCountOfAssignedItemsInArea(lNodeArea);
      if (estimateditemcnt > 4) and (MapView.Zoom < MapView.ZoomMax) then
      begin
        lItemGroupMarker := TGpsPointGroupMarker.Create((rpt1.Lon+rpt2.Lon)/2,(rpt1.Lat+rpt2.Lat)/2.0);
        lItemGroupMarker.GroupedItemsCount := Round(estimateditemcnt);
        MapView.GPSItems.Add(lItemGroupMarker,0);
      end
      else
      begin
        itemcnt := FTestQuadTree.CountAssignedItemsInArea(lNodeArea);
        if itemcnt > 0 then
        begin
          lList := TList.Create;
          try
            FTestQuadTree.ListItemsInArea(lNodeArea,lList);
            for i := 0 to lList.Count-1 do
            begin
              lpoi := TTestPOI(lList.Items[i]);
              if Assigned(lpoi) then
              begin
                gpsPt := TGpsPoint.Create(lpoi.Lon,lpoi.Lat);
                gpsPt.Name := lpoi.Caption;
                MapView.GPSItems.Add(gpspt,0);
              end;
            end;
          finally
            lList.Free;
          end;
        end;
      end;
    end;
  end;
  lblListItemsCount.Caption := IntToStr(MapView.GPSItems.Count);
{
  gpsPt := TGpsPoint.Create(MapView.Center.Lon,MapView.Center.Lat);
  MapView.GPSItems.Add(gpspt,0);
}
*)
end;


procedure TForm1.btnAreaHitTestClick(Sender: TObject);
begin
  if not Assigned(FLatLonHitThread) then
  begin
    FLatLonHitThread := TLatLonHitThread.Create(False);
    btnAreaHitTest.Caption := 'Stop AreaHitTest';
  end
  else
  begin
    FreeAndNil(FLatLonHitThread);
    btnAreaHitTest.Caption := 'Start AreaHitTest';
  end;
end;

type
  PCoordEdge = ^TCoordEdge;
  TCoordEdge = record
    Posi : Double; // Location of the Edge
    Direction : Integer; // Direction +1 = rising, -1 = falling
    Saldo : Integer;
  end;
function CoordEdgeListSortCompare(Item1, Item2: Pointer): Integer;
var
  p1 : PCoordEdge absolute Item1;
  p2 : PCoordEdge absolute Item2;
begin
  if p1^.Posi > p2^.Posi then
    Result := 1
  else if p1^.Posi < p2^.Posi then
    Result := -1
  else if p1^.Direction < p2^.Direction then
    Result := -1
  else if p1^.Direction > p2^.Direction then
    Result := 1
  else
    Result := 0;
end;

procedure TForm1.btnRectReliefClick(Sender: TObject);
var
  ce : array of TCoordEdge;
  i : Integer;
  ls : TList;
  p : PCoordEdge;
  c : TCoordEdge;
  saldo : Integer;
  qn : TLazQuadNode;
  o : TObject;
  ndx : Integer;
  shp : TTestShape;
begin
  SetLength(ce,0);
  qn := FTestQuadTree.RootNode;
  while Assigned(qn) do
  begin
    for i := 0 to qn.NodeItemsLength-1 do
    begin
      o := qn.NodeItems[i];
      if (not Assigned(o)) or (not (o is TTestShape)) then Continue;
      shp := TTestShape(o);
      if shp.ShapeBounds.Left > -35 then
      begin
        ndx := Length(ce);
        SetLength(ce,Length(ce)+2);
        ce[ndx].Posi:= shp.ShapeBounds.Bottom; //shp.ShapeBounds.Left;
        ce[ndx].Direction:= 1;
        ce[ndx+1].Posi:= shp.ShapeBounds.Top; //shp.ShapeBounds.Right;
        ce[ndx+1].Direction:= -1;
      end;
    end;
    qn := FTestQuadTree.NextNode(qn);
  end;

{
  SetLength(ce,6);

  ce[0].Posi := 0.0;
  ce[0].Direction := 1;
  ce[1].Posi := 10.0;
  ce[1].Direction := -1;

  ce[2].Posi := -180.0;
  ce[2].Direction := 1;
  ce[3].Posi := 0.0;
  ce[3].Direction := -1;


  ce[4].Posi := -150.0;
  ce[4].Direction := 1;
  ce[5].Posi := 0.0;
  ce[5].Direction := -1;
}

  ls := TList.Create;
  try
    for i := 0 to High(ce) do
      ls.Add(@ce[i]);
    ls.Sort(@CoordEdgeListSortCompare);
    c := PCoordEdge(ls.Items[0])^;
    Saldo := c.Direction;
    for i := 1 to ls.Count-1 do
    begin
      p := PCoordEdge(ls.Items[i]);
      if (p^.Posi <> c.Posi) or
         ((p^.Direction > 0) and (c.Direction < 0)) or
         ((p^.Direction < 0) and (c.Direction > 0)) then
      begin
        Memo3.Lines.Add(Format('%1.2f%s%d%s%d',[c.Posi,#9, c.Direction,#9, Saldo]));
        c := p^;
      end
      else
      begin
        c.Direction := c.Direction + p^.Direction;
      end;
      Saldo := Saldo + p^.Direction;
    end;
    Memo3.Lines.Add(Format('%1.2f%s%d%s%d',[c.Posi,#9, c.Direction, #9, Saldo]));
  finally
    ls.Free;
  end;


end;

procedure TForm1.btnSampleImportClick(Sender: TObject);
var
  sl, line : TStringList;
  i : Integer;
  llat, llon : Double;
  lpoi : TTestPOI;
  lineerr, cnt0, cnt1 : Integer;
  ini : TMemIniFile;
  keys : TStringList;
  shp : TTestShape;
  j : Integer;
  cnt : Integer;
  shpstr : String;
  gpsPt : TGpsPoint;
begin
  if (not cbPOISample.Checked) and
     (not cbShapeSample.Checked) then Exit;

  if cbPOIInsertLazMap.Checked then
    MapView.GPSItems.Clear(0);

  FTestQuadTree.Clear;
  FLastMapArea.Init(0,0,0,0); // Force Repaint
//  FTestQuadTree.BeginUpdate;
  try
    if cbPOISample.Checked then
    begin
      line := Nil;
      sl := Nil;
      try
        sl := TStringList.Create;
        line := TStringList.Create;
        line.Delimiter:=',';
        line.StrictDelimiter := True;
        line.QuoteChar:='"';
        sl.LoadFromFile('./POI.csv');
        cnt1 := 0;
        cnt0 := sl.Count;
        lineerr := -1;
        for i := 0 to sl.Count-1 do
        begin
//          if i > 50 then Break; // zum Debuggen, raus hier

          line.DelimitedText := sl[i];
          if line.Count <> 3 then Continue;
          llon := StrToFloatDef(line[0],-181);
          llat := StrToFloatDef(line[1],-181);
          if (llat < -90) or (llat > 90) or
             (llon < -180) or (llon > 180) then
          begin
            if lineerr < 0 then
              lineerr := i;
            Continue;
          end;
          // Einfügen in den MapView
          // dauert ewig :-)
          if cbPOIInsertLazMap.Checked then
          begin
            gpsPt := TGpsPoint.Create(llon,llat);
            MapView.GPSItems.Add(gpspt,0);
          end;

          lpoi := TTestPOI.Create(llat,llon,line[2]);
          if not Assigned(FTestQuadTree.InsertItem(lpoi)) then
          begin
            if lineerr < 0 then
              lineerr := i;
            FreeAndNil(lpoi);
          end
          else
          begin
            lpoi := Nil;
            Inc(cnt1);
          end;
        end;
        if lineerr >= 0 then
          lblSampleImportResult.Caption := Format('%d lines imported into %d objects. Error in Line %d',[cnt0,cnt1,lineerr])
        else
          lblSampleImportResult.Caption := Format('%d lines imported into %d objects',[cnt0,cnt1]);

    //    sl.Clear;
    //    FTestQuadTree.TreeAsStrings(sl);
    //    sl.SaveToFile('./poitree.txt');

      finally
        if Assigned(line) then
          line.Free;
        if Assigned(sl) then
          sl.Free;
      end;
    end;
    if cbShapeSample.Checked then
    begin
      lineerr := 0;
      cnt := 0;
      FCountryShapes.Clear;
      ini := TMemIniFile.Create('.\CountryShapes.ini');
      try
        sl := Nil;
        keys := Nil;
        try
          sl := TStringList.Create;
          keys := TStringList.Create;
          ini.ReadSections(sl);
          for i := 0 to sl.Count-1 do
          begin
            ini.ReadSection(sl[i],keys);
            for j := 0 to keys.Count-1 do
            begin
              shpstr := ini.ReadString(sl[i],keys[j],'');
              if Length(shpstr) < 1 then Continue;
              shp := TTestShape.Create(0,0,sl[i]);
              try
                shp.ReadShapesFromString(shpstr);
                if shp.CoordsCount > 0 then
                begin
                  Memo2.Lines.Add(Format('Loaded "%s" with %d coords',[shp.Caption,shp.CoordsCount]));
                  // FCountryShapes.Add(shp);
                  if not Assigned(FTestQuadTree.InsertItem(shp)) then
                  begin
                    if lineerr < 0 then
                      lineerr := i;
                    Memo2.Lines.Add(Format('Error in section "%s" in key "%s"',[sl[i],keys[j]]));
                    FreeAndNil(shp);
                  end
                  else
                  begin
                    shp := Nil;
                    Inc(cnt);
                  end;
                end;
              finally
                if Assigned(shp) then
                  FreeAndNil(shp);
              end;
            end;
          end;
        finally
          if Assigned(sl) then
            sl.Free;
          if Assigned(keys) then
            keys.Free;
        end;
      finally
        ini.Free;
      end;
      Memo2.Lines.Add(Format('Loaded %d shapes',[cnt]));
    end;
  finally
//    FTestQuadTree.EndUpdate;
  end;
end;

procedure TForm1.btnPolyHitTestClick(Sender: TObject);
var
  poly : array of Double;
  rc : Integer;
  b : Boolean;
//  ax,ay : Double;
  s : String;
  i : Integer;
begin
  SetLength(poly,5*2);
  //           X               Y
  poly[0] :=  0.0; poly[1] :=   0.0;
  poly[2] := 10.0; poly[3] :=   0.0;
  poly[4] := 10.0; poly[5] :=  -10.0;
  poly[6] :=  0.0; poly[7] :=  -10.0;
  poly[8] :=  0.0; poly[9] :=   0.0;
//  ax := 0.0;
//  ay := 0.0;
  for i := 0 to (Length(poly) div 2)-2 do
  begin
    rc := WindingPointInPolygon(poly[i*2],poly[i*2+1],
                                poly[0],
                                Length(poly) div 2,
                                False, // First value is not Y (its X)
                                True); // The Y value increases going up
    b := CrossingNumberPointInPolygon(poly[i*2],poly[i*2+1],
                                poly[0],
                                Length(poly) div 2,
                                False, // First value is not Y (its X)
                                True); // The Y value increases going up
    if b then
      s := 'Crossing result: Inside'
    else
      s := 'Crossing result: Outside';
    Memo3.Lines.Add(Format('X %1.2f, Y %1.2f: %s , WindingResult: %d',[poly[i*2],poly[i*2+1],s,rc]));
  end;
end;

procedure TForm1.btnTreeAsStringsClick(Sender: TObject);
var
  sl : TStringList;
begin
  sl := TStringList.Create;
  try
    FTestQuadTree.TreeAsStrings(sl);
    Memo3.Lines.Assign(sl);
  finally
    sl.Free;
  end;
end;
{ CalculateMapPixelCoords
  calculates the coordinates of the displayed map.
  This coordinates are adjusted to the full tiles, that are necesarry
  to cover the displayed map.
  This means:
  On a small zoom level (approx. < 5) than the full earth is returned.
  If the map is in the Cyclic mode, depending of the center position,
  the left-Coordinate may be negative, indicating that the map extends to the
  eastern Hemisphere. If the right coordinate is larger than the MapSize, this
  indicates that the map extends to the western hemishere.
  Example: If the procedure return
  -256,0,256,512
  than full earth is shown in Zoom-Level 1, the date border is in the middle, subsequently
  the eastern hemisphere is displayed left, the western hemisphere is displayed right.
}
procedure TForm1.CalculateMapPixelCoords(var ALeft, ATop, ARight,
  ABottom: Int64; var AOSMMapSize: Int64; var AOfsX, AOfsY: Integer);
var
  pt0 : TPoint;
  osmapsz : Int64;
  wpx,hpx : Int64;
  minx, maxx, miny, maxy : Int64;
  mcl : Double;
  d : Double;
begin
  wpx := MapView.Width;
  hpx := MapView.Height;
  osmapsz := round(mvGeoMath.ZoomFactor(MapView.Zoom)) * TILE_SIZE;
  AOSMMapSize := osmapsz;
  mcl := MapView.Center.Lon;
  // Pixelkoordinaten der Karte im Zentrum des Fensters. Liegt immer innerhalb der Karte!
  pt0 := DegreesToPixelsOSM(MapView.Zoom,MapView.Center.Lat,mcl);
  // den Offset zur nächsten linken bzw oberen Kante der das Zentrum enthaltenden Kachel berechnen
  AOfsX := (pt0.X mod TILE_SIZE);
  AOfsY := (pt0.Y mod TILE_SIZE);

  // Die Anzahl der ganzen Kacheln zum linken Fensterrand bestimmen
  d := ((wpx / 2)-AOfsX) / TILE_SIZE;
  minx := Trunc(d);
  if Frac(d) > 0 then
    Inc(minx);
  // Daraus die Pixelkoordinate der entsprechenden Kachel
  minx := minx * TILE_SIZE;
  minx := pt0.X - AOfsX - minx;

  // Die Anzahl der ganzen Kacheln zum rechten Fensterrand bestimmen
  d := ((wpx / 2)+AOfsX) / TILE_SIZE;
  maxx := Trunc(d);
  if Frac(d) > 0 then
    Inc(maxx);
  maxx := maxx * TILE_SIZE;
  maxx := pt0.X - AOfsX + maxx;

  // Abfangen der Sonderfälle
  // Falls die Karte nicht zyklisch ist, also an der Datumsgrenze endet
  if not MapView.Cyclic then
  begin
    if minx < 0 then
      minx := 0;
    if maxx > osmapsz then
      maxx := osmapsz;
  end;
  // In allen Fällen, falls die Breite des Fensters größer als die Karte ist
  if (maxx-minx) > osmapsz then
  begin
    minx := 0;
    maxx := osmapsz;
  end;

  // Das selber Spiel mit der Höhe der Karte
  d := ((hpx / 2)-AOfsY) / TILE_SIZE;
  miny := Trunc(d);
  if Frac(d) > 0 then
    Inc(miny);
  miny := miny * TILE_SIZE;
  miny := pt0.Y - AOfsY - miny;

  d := ((hpx / 2)+AOfsY) / TILE_SIZE;
  maxy := Trunc(d);
  if Frac(d) > 0 then
    Inc(maxy);
  maxy := maxy * TILE_SIZE;
  maxy := pt0.Y - AOfsY + maxy;

  // Auch hier die absoluten Grenzen festlegen
  if (maxy-miny) > osmapsz then
  begin
    miny := 0;
    maxy := osmapsz;
  end;
  if miny < 0 then
    miny := 0;
  if maxy > osmapsz then
    maxy := osmapsz;


(*
  // Die Pixelkoorndinate der oberen linken Kante der Kachel bestimmen in der das Zentrum liegt.
  pt1.X := pt0.X-ofsx;
  pt1.Y := pt0.Y-ofsy;
  // Die X-Koordinaten rechts und links bestimmen
  minx := pt1.X;
  // so oft nach links gehen, bis das Fenster verlassen wird
  while (minx > (pt0.X - (wpx div 2))) do
    minx := minx - TILE_SIZE;

  maxx := pt1.X;
  // so oft nach rechts gehen, das Fenster verlassen wird
  while (maxx < (pt0.X + (wpx div 2))) do
    maxx := maxx + TILE_SIZE;

  if not MapView.Cyclic then
  begin  // Fall, dass die Karte nicht über die Datumsgrenze hinausläuft.
    if minx < 0 then
      minx := 0;
    if maxx > osmapsz then
      maxx := osmapsz;
  end
  else // Sonderfall, zyklische Karte mit Datumsgrenze
  begin
    winmapsz := maxx-minx;
    if winmapsz > osmapsz then // Das Fenster ist breiter als die OSM-Karte
    begin
      minx := 0;
      maxx := osmapsz;
    end;
{
    else if dtlinecr then
    begin

    // Eine der beiden X Werte ist am Poller. Rechts oder links ist noch Karte vorhanden.
    if (mcl <> 0) and // Die Zentrums-Longitude ist ungleich 0
       ((minx = 0) or (maxx = osmapsz)) then
    begin
      if (minx = 0) then
        minx := maxx - osmapsz
      else if (maxx = osmapsz) then
        maxx := minx + osmapsz
    end;
}
  end;

  // Die Y-Koordinaten bestimmen
  miny := pt1.Y;
  // so oft nach oben gehen, bis entweder die Karte zu Ende ist, oder das Fenster verlassen wird
  while (miny > 0) and (miny > (pt0.Y - (hpx div 2))) do
    miny := miny - TILE_SIZE;

  maxy := pt1.Y;
  // so oft nach rechts gehen, bis entweder die Karte zu Ende ist, oder das Fenster verlassen wird
  while (maxy < osmapsz) and (maxy < (pt0.Y + (hpx div 2))) do
    maxy := maxy + TILE_SIZE;
*)

  // Das Ergebnis zuweisen
  ALeft := minx;
  ARight := maxx;
  ATop := miny;
  ABottom := maxy;
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  lLeft,lTop,lRight,lBottom : Int64;
  lOSMMapSize : Int64;
  x0,x1 : Int64;
  lLat0, lLon0, lLat1, lLon1 : Double;
  ofsx, ofsy : Integer;
begin
  lLeft := 0;
  lTop := 0;
  lRight := 0;
  lBottom := 0;
  lOSMMapSize := 0;
  CalculateMapPixelCoords(lLeft,lTop,lRight,lBottom,lOSMMapSize,ofsx,ofsy);
  x0 := lLeft;
  if x0 < 0 then
    x0 := lOSMMapSize+x0;
  x1 := lRight;
  if x1 > lOSMMapSize then
    x1 := x1-lOSMMapSize;
  PixelsToDegreesOSM(x0,lTop,MapView.Zoom,lLat0,lLon0);
  PixelsToDegreesOSM(x1,lBottom,MapView.Zoom,lLat1,lLon1);
  Memo4.Text := Format('MapXY %d,%d; %d,%d',[lLeft,lTop,lRight,lBottom]);
  Memo4.Lines.Add(Format('MapW %d, OfsX,Y %d, %d',[lOSMMapSize,ofsx,ofsy]));
  Memo4.Lines.Add(Format('Lat/Lon %1.1f, %1.1f; %1.1f, %1.1f',[lLat0,lLon0, lLat1, lLon1]));
end;

{
var
  pt0, pt1 : TPoint;
  ofsx,ofsy : Integer;
  osmapsz : Int64;
  wpx,hpx : Integer;
  dtlinecr : Boolean;
  hmappx, wmappx : Integer;
  minx, maxx, miny, maxy : Integer;
  mcl : Double;
  lLat0, lLon0, lLat1, lLon1 : Double;
  x0,x1,y0,y1 : Integer;
begin
  wpx := MapView.Width;
  hpx := MapView.Height;
  osmapsz := round(mvGeoMath.ZoomFactor(MapView.Zoom)) * TILE_SIZE;
  dtlinecr := MapView.Engine.CrossesDateline;
  // Wenn die OSM Karte schmaler als das Fenster ist, dann die Kartenmaße, sonst das Fenster
  if osmapsz < wpx then
    wmappx := osmapsz
  else
    wmappx := wpx;
  // Wenn die OSM Karte kürzer als das Fenster ist, dann die Kartenmaße, sonst das Fenster
  if osmapsz < hpx then
    hmappx := osmapsz
  else
    hmappx := hpx;

  mcl := MapView.Center.Lon;
  // Pixelkoordinaten der Karte im Zentrum des Fensters. Liegt immer innerhalb der Karte!
  pt0 := DegreesToPixelsOSM(MapView.Zoom,MapView.Center.Lat,mcl);
  // den Offset zur nächsten linken bzw oberen Kante der das Zentrum enthaltenden Kachel berechnen
  ofsx := (pt0.X mod TILE_SIZE);
  ofsy := (pt0.Y mod TILE_SIZE);

  // Die Pixelkoorndinate der oberen linken Kante der Kachel bestimmen.
  pt1.X := pt0.X-ofsx;
  pt1.Y := pt0.Y-ofsy;

  minx := pt1.X;
  // so oft nach links gehen, bis entweder die Karte zu Ende ist, oder das Fenster verlassen wird
  while (minx > 0) and (minx > (pt0.X - (wpx div 2))) do
    minx := minx - TILE_SIZE;

  maxx := pt1.X;
  // so oft nach rechts gehen, bis entweder die Karte zu Ende ist, oder das Fenster verlassen wird
  while (maxx < osmapsz) and (maxx < (pt0.X + (wpx div 2))) do
    maxx := maxx + TILE_SIZE;

  // Sonderfall, zyklische Karte mit Datumsgrenze
  if dtlinecr then
  begin
    // Eine der beiden X Werte ist am Poller der Karte und rechts oder links ist noch Karte vorhanden
    if (minx = 0) or
       (maxx = osmapsz) then
    begin
      if (mcl > 0) then
      begin
        if (minx = 0) then
          minx := maxx - osmapsz
        else if (maxx = osmapsz) then
          maxx := minx + osmapsz
      end
      else if (mcl < 0) then
      begin
        if minx = 0 then
          minx := maxx - osmapsz
        else if (maxx = osmapsz) then
          maxx := minx + osmapsz;
      end;
    end;
  end;

  miny := pt1.Y;
  // so oft nach oben gehen, bis entweder die Karte zu Ende ist, oder das Fenster verlassen wird
  while (miny > 0) and (miny > (pt0.Y - (hpx div 2))) do
    miny := miny - TILE_SIZE;

  maxy := pt1.Y;
  // so oft nach rechts gehen, bis entweder die Karte zu Ende ist, oder das Fenster verlassen wird
  while (maxy < osmapsz) and (maxy < (pt0.Y + (hpx div 2))) do
    maxy := maxy + TILE_SIZE;

  x0 := minx;
  if x0 < 0 then
    x0 := minx+osmapsz;
  x1 := maxx;
  if x1 > osmapsz then
    x1 := maxx-osmapsz;
  PixelsToDegreesOSM(x0,miny,MapView.Zoom,lLat0,lLon0);
  PixelsToDegreesOSM(x1,maxy,MapView.Zoom,lLat1,lLon1);
  Memo4.Text := Format('MapXY %d,%d; %d,%d',[minx,miny,maxx,maxy]);
  Memo4.Lines.Add(Format('Lat/Lon %1.1f, %1.1f; %1.1f, %1.1f',[lLat0,lLon0, lLat1, lLon1]))


end;
}

procedure TForm1.FormActivate(Sender: TObject);
begin
  OnActivate := Nil;
  MapView.Active := True;
  MapView.ZoomToCursor:= True;;
end;

procedure TForm1.FormCreate(Sender: TObject);
var
  fn : String;
  ws : TLazQuadNodeArea;
begin
  DefaultFormatSettings.DecimalSeparator:='.';
  DefaultFormatSettings.ThousandSeparator:=#0;
  FWorldmap := TBitmap.Create;
  fn := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName))+'EarthMap.bmp';
  FWorldmap.LoadFromFile(fn);
  ws.Left := -180.0;
  ws.Top := 90.0;
  ws.Right := 180.0;
  ws.Bottom := -90.0;
  FTestQuadTree := TOSMQuadTree.Create(ws,True);
  FTestQuadTree.MaxQuadTreeLevel := MapView.ZoomMax; //Equal Tree and Map-Zoom levels
  FTestQuadTree.OnGetItemCoord := @OnTestQuadTreeGetItemCoord;
  FTestQuadTree.OnGetItemArea := @OnTestQuadTreeGetItemAreaEvent;
  FTestQuadTree.OnGetItemCaption := @OnTestQuadTreeGetItemCaptionEvent;
  FBaseRect.Left := -100;
  FBaseRect.Top := 20;
  FBaseRect.Right := 100;
  FBaseRect.Bottom := -20;
  FTestRect.Left := -20;
  FTestRect.Top := 40;
  FTestRect.Right := 20;
  FTestRect.Bottom := -40;

  FCountryShapes := TObjectList.Create(True);
  FLastMapArea.Init(0,0,0,0);
  MapView.OnBeforeDrawObjects:=@OnMapViewBeforeDrawObjects;

end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FTestQuadTree.Free;
  FWorldmap.Free;

  FCountryShapes.Free;

end;

procedure TForm1.MapViewChange(Sender: TObject);
var
  s : String;
begin
  lblPOIDesc.Caption := '';
end;

procedure TForm1.MapViewDrawGpsPoint(Sender: TObject;
  ADrawer: TMvCustomDrawingEngine; APoint: TGpsPoint);
var
  P: TPoint;
  Pt: TPoint;
  PtCyc: TPointArray;
  s : String;
  d: Integer;
//  sym: TGPSSymbol;
  clr: TColor;
  extent: TSize;
  lMapView: TMapView;
  cnt : Integer;
begin
  lMapView := TMapView(Sender);

  // Screen coordinates of the GPS point
  P := lMapView.LatLonToScreen(APoint.RealPoint);

  PtCyc := MapView.CyclicPointsOf(P);
  //  sym := gpsPlus;
  clr := clRed;
  d := 5;
  ADrawer.PenColor := clr;
  ADrawer.PenWidth:= 3;
  for Pt in PtCyc do
  begin
    if APoint is TGpsPointGroupMarker then
    begin
      ADrawer.PenColor := clNavy;
      ADrawer.BrushStyle := bsSolid;
      ADrawer.BrushColor := clWhite;
      d := 10;
      ADrawer.Ellipse(Pt.X - d, Pt.Y - d,
                      Pt.X + d, Pt.Y + d);
      cnt := TGpsPointGroupMarker(APoint).GroupedItemsCount;
      if cnt > 9999 then
        s := '10K+'
      else
        s := IntToStr(cnt);
      extent := ADrawer.TextExtent(s);
      // ... and font
      ADrawer.BrushStyle := bsClear;
      ADrawer.FontColor := clBlack;
      ADrawer.FontName := 'Arial';
      ADrawer.FontSize := 6;
      ADrawer.FontStyle := [];

      // Write the POI text
      ADrawer.TextOut(Pt.X - (extent.CX div 2), Pt.Y - (extent.CY div 2), s);

    end
    else
    begin
      ADrawer.Line(Pt.X - d, Pt.Y, Pt.X + d, Pt.Y);
      ADrawer.Line(Pt.X, Pt.Y - d, Pt.X , Pt.Y + d);
{
      // Prepare text output: background color...
    //  inc(Pt.Y, d + 4);
      extent := ADrawer.TextExtent(APoint.Name);
      ADrawer.BrushStyle := bsClear;
      // ... and font
      ADrawer.FontColor := MapView.Font.Color;
      ADrawer.FontName := MapView.Font.Name;
      ADrawer.FontSize := MapView.Font.Size;
      ADrawer.FontStyle := MapView.Font.Style;


      // Write the POI text
      ADrawer.TextOut(Pt.X - extent.CX div 2, Pt.Y + d + 4, APoint.Name);
}
    end;
  end;
{
  // Draw the GPS point as specified by the data in the point's ExtraData
  if (APoint.ExtraData is TGPSExtraData) then
    // Get the POI attributes
    with TGPSExtraData(APoint.ExtraData) do
    begin
      clr := Color;
      sym := Symbol;
      d := Size div 2;
    end
  else
  begin
    sym := gpsPlus;
    clr := clRed;
    d := 5;
  end;


  // Draw the POI symbol
  ADrawer.PenColor := clr;
  case sym of
    gpsPlus:
      begin
        ADrawer.Line(P.X - d, P.Y, P.X + d, P.Y);
        ADrawer.Line(P.X, P.Y - d, P.X , P.Y + d);
      end;
    gpsCross:
      begin
        ADrawer.Line(P.x - d, P.Y - d, P.X + d, P.Y + d);
        ADrawer.Line(P.x - d, P.Y + d, P.X + d, P.Y - d);
      end;
    gpsFilledCircle:
      begin
        ADrawer.BrushStyle := bsSolid;
        ADrawer.BrushColor := clr;
        ADrawer.Ellipse(P.X - d, P.Y - d, P.X + d, P.Y + d);
      end;
    gpsOpenCircle:
      begin
        ADrawer.BrushStyle := bsClear;
        ADrawer.Ellipse(P.X - d, P.Y - d, P.X + d, P.Y + d);
      end;
    gpsFilledRect:
      begin
        ADrawer.BrushStyle := bsSolid;
        ADrawer.BrushColor := clr;
        ADrawer.Rectangle(P.X - d, P.Y - d, P.X + d, P.Y + d);
      end;
    gpsOpenRect:
      begin
        ADrawer.BrushStyle := bsClear;
        ADrawer.Rectangle(P.X - d, P.Y - d, P.X + d, P.Y + d);
      end;
  end;

  // Prepare text output: background color...
  inc(P.Y, d + 4);
  extent := ADrawer.TextExtent(APoint.Name);
  if CfgFrame.POITextBkColor = clNone then
    ADrawer.BrushStyle := bsClear
  else
  begin
    ADrawer.BrushStyle := bsSolid;
    ADrawer.BrushColor := CfgFrame.POITextBkColor;
  end;

  // ... and font
  ADrawer.FontColor := MapView.Font.Color;
  ADrawer.FontName := MapView.Font.Name;
  ADrawer.FontSize := MapView.Font.Size;
  ADrawer.FontStyle := MapView.Font.Style;

  // Write the POI text
  ADrawer.TextOut(P.X - extent.CX div 2, P.Y, APoint.Name);
}
end;

procedure TForm1.MapViewMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FMapViewMouseDownX := X;
  FMapViewMouseDownY := Y;
end;

procedure TForm1.MapViewMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
const
  DELTA = 3;
var
  rArea: TRealArea;
  gpsList: TGpsObjList;
  sl : TStringList;
  i: Integer;
  pt : TPoint;
begin
  pt.X := X-DELTA;
  pt.Y := Y-DELTA;
  rArea.TopLeft := MapView.ScreenToLatLon(pt);
  pt.X := X+DELTA;
  pt.Y := Y+DELTA;
  rArea.BottomRight := MapView.ScreenToLatLon(pt);
  gpsList := MapView.GpsItems.GetObjectsInArea(rArea);
  try
    if gpsList.Count > 0 then
    begin
      sl := TStringList.Create;
      try
        for i:=0 to gpsList.Count-1 do
        begin
          if (gpsList[i] is TGpsPoint) and
             (not (gpsList[i] is TGpsPointGroupMarker)) then
            sl.Add(TGpsPoint(gpsList[i]).Name);
        end;
        lblPOIDesc.Caption := sl.Text;
      finally
        sl.Free;
      end;
    end
    else
      lblPOIDesc.Caption := '';
  finally
    gpsList.Free;
  end;
end;

procedure TForm1.MapViewMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
const
  DELTA = 12;
var
  rArea: TRealArea;
  gpsList: TGpsObjList;
  L: TStrings;
  i: Integer;
  pt : TPoint;
  rpt : TRealPoint;
begin
  if (Abs(FMapViewMouseDownX-X) > 2) or
     (Abs(FMapViewMouseDownY-Y) > 2) then Exit;

  pt.X := X-DELTA;
  pt.Y := Y-DELTA;
  rArea.TopLeft := MapView.ScreenToLatLon(pt);
  pt.X := X+DELTA;
  pt.Y := Y+DELTA;
  rArea.BottomRight := MapView.ScreenToLatLon(pt);
  gpsList := MapView.GpsItems.GetObjectsInArea(rArea);
  try
    if gpsList.Count > 0 then
    begin
      L := TStringList.Create;
      try
        for i:=0 to gpsList.Count-1 do
        begin
          if gpsList[i] is TGpsPointGroupMarker then
          begin
            MapView.Center := TGpsPointGroupMarker(gpsList[i]).RealPoint;
            MapView.Zoom := MapView.Zoom+3;
          end;
        end;
      finally
        L.Free;
      end;
    end;
  finally
    gpsList.Free;
  end;
end;

procedure TForm1.pbEarthMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FMouseIsDown := True;
  FMouseDownHit := PaintMouseHitTest(X,Y,Shift);
  case FMouseDownHit of
    mrhTestArea,
    mrhTestTop, mrhTestBottom, mrhTestRight, mrhTestLeft,
    mrhTestNW, mrhTestNE, mrhTestSE, mrhTestSW :
      FMouseDownLatLonRect := FTestRect;
    mrhBaseArea,
    mrhBaseTop, mrhBaseBottom, mrhBaseRight, mrhBaseLeft,
    mrhBaseNW, mrhBaseNE, mrhBaseSE, mrhBaseSW :
      FMouseDownLatLonRect := FBaseRect;
  end;
  FMouseDownLastX := X;
  FMouseDownLastY := Y;

end;

procedure TForm1.pbEarthMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  llat, llon : Double;
  DeltaX, DeltaY : Integer;
  h : TMouseRectHit;
{
  = (mrhNone,
                 mrhTestArea,
                 mrhTestTop, mrhTestBottom, mrhTestRight, mrhTestLeft,
                 mrTestNW, mrTestNE, mrTestSE, mrTestSW,
                 mrhBaseArea,
                 mrhBaseTop, mrhBaseBottom, mrhBaseRight, mrhBaseLeft,
                 mrBaseNW, mrBaseNE, mrBaseSE, mrBaseSW
                 );
}
begin
  llat := -(Y / 2)+90;
  llon := (X /2)-180;
  if llon < -180 then
    llon := -180
  else if llon > 180 then
    llon := 180;
  sbMapStatus.Panels[1].Text := Format('%1.1f',[llat]);
  sbMapStatus.Panels[3].Text := Format('%1.1f',[llon]);

  if not FMouseIsDown then
  begin

    h := PaintMouseHitTest(X,Y,Shift);
    case h of
      mrhTestArea,mrhBaseArea : pbEarth.Cursor := crSize;
      mrhTestTop, mrhTestBottom,
      mrhBaseTop, mrhBaseBottom : pbEarth.Cursor := crSizeNS;
      mrhTestRight, mrhTestLeft,
      mrhBaseRight, mrhBaseLeft : pbEarth.Cursor := crSizeWE;
      mrhTestNW, mrhTestSE,
      mrhBaseNW, mrhBaseSE: pbEarth.Cursor := crSizeNWSE;
      mrhTestNE,  mrhTestSW,
      mrhBaseNE, mrhBaseSW : pbEarth.Cursor := crSizeNESW;
    else
      pbEarth.Cursor := crDefault;
    end;

  end
  else // Mouse is Down
  begin
    if FMouseDownHit = mrhNone then Exit;
    DeltaX := X-FMouseDownLastX;
    DeltaY := Y-FMouseDownLastY;
    case FMouseDownHit of
      mrhTestArea :
        begin
          FTestRect.Left  := FMouseDownLatLonRect.Left + DeltaX*0.5;
          FTestRect.Right  := FMouseDownLatLonRect.Right + DeltaX*0.5;
          FTestRect.Top := FMouseDownLatLonRect.Top - DeltaY*0.5;
          FTestRect.Bottom := FMouseDownLatLonRect.Bottom - DeltaY*0.5;
        end;
      mrhBaseArea :
        begin
          FBaseRect.Left  := FMouseDownLatLonRect.Left + DeltaX*0.5;
          FBaseRect.Right  := FMouseDownLatLonRect.Right + DeltaX*0.5;
          FBaseRect.Top := FMouseDownLatLonRect.Top - DeltaY*0.5;
          FBaseRect.Bottom := FMouseDownLatLonRect.Bottom - DeltaY*0.5;
        end;
      mrhTestTop :
        begin
          FTestRect.Top := FMouseDownLatLonRect.Top - DeltaY*0.5;
        end;
      mrhTestBottom :
        begin
          FTestRect.Bottom := FMouseDownLatLonRect.Bottom - DeltaY*0.5;
        end;
      mrhBaseTop :
        begin
          FBaseRect.Top := FMouseDownLatLonRect.Top - DeltaY*0.5;
        end;
      mrhBaseBottom :
        begin
          FBaseRect.Bottom := FMouseDownLatLonRect.Bottom - DeltaY*0.5;
        end;
      mrhTestRight :
        begin
          FTestRect.Right := FMouseDownLatLonRect.Right + DeltaX*0.5;
        end;
      mrhTestLeft :
        begin
          FTestRect.Left := FMouseDownLatLonRect.Left + DeltaX*0.5;
        end;
      mrhBaseRight:
        begin
          FBaseRect.Right := FMouseDownLatLonRect.Right + DeltaX*0.5;
        end;
      mrhBaseLeft:
        begin
          FBaseRect.Left := FMouseDownLatLonRect.Left + DeltaX*0.5;
        end;
      mrhTestNW:
        begin
          FTestRect.Left  := FMouseDownLatLonRect.Left + DeltaX*0.5;
          FTestRect.Top := FMouseDownLatLonRect.Top - DeltaY*0.5;
        end;
      mrhTestSE:
        begin
          FTestRect.Right  := FMouseDownLatLonRect.Right + DeltaX*0.5;
          FTestRect.Bottom := FMouseDownLatLonRect.Bottom - DeltaY*0.5;
        end;
      mrhBaseNW:
        begin
          FBaseRect.Left  := FMouseDownLatLonRect.Left + DeltaX*0.5;
          FBaseRect.Top := FMouseDownLatLonRect.Top - DeltaY*0.5;
        end;
      mrhBaseSE:
        begin
          FBaseRect.Right  := FMouseDownLatLonRect.Right + DeltaX*0.5;
          FBaseRect.Bottom := FMouseDownLatLonRect.Bottom - DeltaY*0.5;
        end;
      mrhTestNE:
        begin
          FTestRect.Right  := FMouseDownLatLonRect.Right + DeltaX*0.5;
          FTestRect.Top := FMouseDownLatLonRect.Top - DeltaY*0.5;
        end;
      mrhTestSW:
        begin
          FTestRect.Left  := FMouseDownLatLonRect.Left + DeltaX*0.5;
          FTestRect.Bottom := FMouseDownLatLonRect.Bottom - DeltaY*0.5;
        end;
      mrhBaseNE:
        begin
          FBaseRect.Right  := FMouseDownLatLonRect.Right + DeltaX*0.5;
          FBaseRect.Top := FMouseDownLatLonRect.Top - DeltaY*0.5;
        end;
      mrhBaseSW:
        begin
          FBaseRect.Left  := FMouseDownLatLonRect.Left + DeltaX*0.5;
          FBaseRect.Bottom := FMouseDownLatLonRect.Bottom - DeltaY*0.5;
        end;
    end;
    pbEarth.Invalidate;
    pbEarth.Repaint;
  end;
end;

procedure TForm1.pbEarthMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  d : Double;
begin
  FMouseIsDown := False;
  pbEarth.Cursor := crDefault;

  case FMouseDownHit of
    mrhTestArea,
    mrhTestTop, mrhTestBottom, mrhTestRight, mrhTestLeft,
    mrhTestNW, mrhTestNE, mrhTestSE, mrhTestSW :
      begin
        NormalizeAndUnfoldArea(FTestRect,qydUpPlus,True,FTestQuadTree.RootNode.NodeArea);
        while FTestRect.Left < -180 do
          FTestRect.Left := FTestRect.Left + 360;
        while FTestRect.Left > 180 do
          FTestRect.Left := FTestRect.Left - 360;
        while FTestRect.Right < -180 do
          FTestRect.Right := FTestRect.Right + 360;
        while FTestRect.Right > 180 do
          FTestRect.Right := FTestRect.Right - 360;
        LatLonRectToGUI;
        d := FTestQuadTree.EstimatedCountOfAssignedItemsInArea(FTestRect);
//        Memo3.Lines.Assign(FTestQuadTree.DebugList);
        sbMapStatus.Panels[5].Text := Format('%1.1f',[d]);
      end;
    mrhBaseArea,
    mrhBaseTop, mrhBaseBottom, mrhBaseRight, mrhBaseLeft,
    mrhBaseNW, mrhBaseNE, mrhBaseSE, mrhBaseSW :
      begin
        NormalizeAndUnfoldArea(FBaseRect,qydUpPlus,True,FTestQuadTree.RootNode.NodeArea);
        while FBaseRect.Left < -180 do
          FBaseRect.Left := FBaseRect.Left + 360;
        while FBaseRect.Left > 180 do
          FBaseRect.Left := FBaseRect.Left - 360;
        while FBaseRect.Right < -180 do
          FBaseRect.Right := FBaseRect.Right + 360;
        while FBaseRect.Right > 180 do
          FBaseRect.Right := FBaseRect.Right - 360;
        LatLonRectToGUI;
      end;
  end;


end;

procedure TForm1.pbEarthPaint(Sender: TObject);
var
  r,rw : TRect;
  rc : Boolean;
  poly : array of TPoint;
  i,k : Integer;
  shp : TTestShape;
  qn : TLazQuadNode;
  o : TObject;
  ndx : Integer;

begin
{
  pbEarth.Canvas.Brush.Style := bsSolid;
  pbEarth.Canvas.Pen.Style := psSolid;
  for i := 0 to High(ColorSet) do
  begin
    pbEarth.Canvas.Brush.Color := ColorSet[i];
    pbEarth.Canvas.Pen.Color := ColorSet[i];
    r.Top := 0;
    r.Bottom := 50;
    r.left := i*25;
    r.Right := r.Left + 25;
    pbEarth.Canvas.Rectangle(r);
  end;

exit;
}
  pbEarth.Canvas.Draw(0,0,FWorldmap);
  pbEarth.Canvas.Brush.Style := bsClear;
  pbEarth.Canvas.Pen.Style := psSolid;
  pbEarth.Canvas.Pen.Color := clLime;
{
  rc := LatLonRectToCanvasRect(FBaseRect,r,rw);
  pbEarth.Canvas.Rectangle(r);
  if rc then
    pbEarth.Canvas.Rectangle(rw);
}

  pbEarth.Canvas.Pen.Color := clRed;
  rc := LatLonRectToCanvasRect(FTestRect,r,rw);
  pbEarth.Canvas.Rectangle(r);
  if rc then
    pbEarth.Canvas.Rectangle(rw);

  ColorEarthMap(pbEarth.Canvas);

  qn := FTestQuadTree.RootNode;
  while Assigned(qn) do
  begin
    ndx := (qn.NodeLevel mod 4);
    if Assigned(qn.ParentNode) then
    begin
      if qn = qn.ParentNode.ChildNodeNE then
        ndx := ndx + 0
      else if qn = qn.ParentNode.ChildNodeSE then
        ndx := ndx + 4
      else if qn = qn.ParentNode.ChildNodeSW then
        ndx := ndx + 8
      else if qn = qn.ParentNode.ChildNodeNW then
        ndx := ndx + 12;
    end;
    for i := 0 to qn.NodeItemsLength-1 do
    begin
      o := qn.NodeItems[i];
      if (not Assigned(o)) or (not (o is TTestShape)) then Continue;
      shp := TTestShape(o);
      SetLength(poly,shp.CoordsCount);
      for k := 0 to High(poly) do
      begin
        poly[k] := LatLonToCavasPoint(shp.CoordsLats[k],shp.CoordsLons[k]);
      end;
      pbEarth.Canvas.Pen.Color := ColorSet[ndx]; //:= clRed;
      pbEarth.Canvas.Polygon(poly);

      pbEarth.Canvas.Pen.Color := clNavy;
      rc := LatLonRectToCanvasRect(shp.ShapeBounds,r,rw);
      r.Right := r.Right +1;
      r.Bottom := r.Bottom +1;
      pbEarth.Canvas.Rectangle(r);
      if rc then
        pbEarth.Canvas.Rectangle(rw);
    end;
    qn := FTestQuadTree.NextNode(qn);
  end;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  if Assigned(FLatLonHitThread) then
  begin
    lblHitTestCounter.Caption := IntToStr(FLatLonHitThread.TestCounter);
    if FLatLonHitThread.Terminated then
    begin
      Memo1.Clear;
      if FLatLonHitThread.HitTestError then
      begin
        Memo1.Lines.Add('Error!');
        Memo1.Lines.Add(Format('BaseArea Left %1.2f, Top %1.2f, Right %1.2f, Bottom %1.2f',
                          [FLatLonHitThread.LatLonRectBase.Left, FLatLonHitThread.LatLonRectBase.Top,
                           FLatLonHitThread.LatLonRectBase.Right, FLatLonHitThread.LatLonRectBase.Bottom]));
        Memo1.Lines.Add(Format('TestArea Left %1.2f, Top %1.2f, Right %1.2f, Bottom %1.2f; ',
                          [FLatLonHitThread.LatLonRectTest.Left,  FLatLonHitThread.LatLonRectTest.Top,
                           FLatLonHitThread.LatLonRectTest.Right, FLatLonHitThread.LatLonRectTest.Bottom]));
        Memo1.Lines.Add(Format('ExpectedHit %s, calculated hit %s',[HitString[FLatLonHitThread.HitExpected],HitString[FLatLonHitThread.HitCalculated]]));
      end
      else
        Memo1.Lines.Add('Test passed!');
      FreeAndNil(FLatLonHitThread);
      btnAreaHitTest.Caption := 'Start AreaHitTest';
    end;
  end;
end;

function TForm1.OnTestQuadTreeGetItemCoord(Sender: TObject; AItem: TObject;
  var AX, AY: Double): Boolean;
begin
  Result := False;
  if Assigned(AItem) then
  begin
    if (AItem is TTestPOI) then
    begin
      AX := TTestPOI(AItem).Lon;
      AY := TTestPOI(AItem).Lat;
      Result := True;
    end;
  end;
end;

function TForm1.OnTestQuadTreeGetItemAreaEvent(Sender: TObject; AItem: TObject;
  var AItemArea: TLazQuadNodeArea): Boolean;
begin
  Result := False;
  if Assigned(AItem) then
  begin
    if (AItem is TTestShape) then
    begin
      AItemArea := TTestShape(AItem).ShapeBounds;
      Result := True;
    end;
  end;
end;

function TForm1.OnTestQuadTreeGetItemCaptionEvent(Sender: TObject;
  AItem: TObject; var AItemCaption: String): Boolean;
begin
  Result := False;
  if Assigned(AItem) then
  begin
    if (AItem is TTestPOI) then
    begin
      AItemCaption := TTestPOI(AItem).Caption;
      Result := True;
    end;
  end;
end;

function TForm1.LatLonToCavasPoint(const ALat, ALon: Double): TPoint;
begin
  Result.X := Round(2*(ALon+180));
  Result.Y := Round(2*(-(ALat-90)));
end;

function TForm1.LatLonRectToCanvasRect(const ALatLonRect: TLazQuadNodeArea;
  var ARect, AWrapRect: TRect): Boolean;
var
  llr: TLazQuadNodeArea;
begin
  Result := False;
  llr := ALatLonRect;
  NormalizeAndUnfoldArea(llr,qydUpPlus,True,TheWorld);
  ARect.Left := Round(2*(llr.Left+180));
  ARect.Top    := Round(2*(-(llr.Top-90)));
  ARect.Right  := Round(2*(llr.Right+180));
  ARect.Bottom := Round(2*(-(llr.Bottom-90)));
  if // (llr.Left < TheWorld.Left) or
     (llr.Right> TheWorld.Right) then
  begin
    Result := True;
    AWrapRect.Left := -1;
    AWrapRect.Top    := Round(2*(-(llr.Top-90)));
    AWrapRect.Right  := Round(2*(llr.Right-360+180));
    AWrapRect.Bottom := Round(2*(-(llr.Bottom-90)));
  end;
end;

procedure TForm1.ColorEarthMap(const ACanvas: TCanvas);
var
  n : TLazQuadNode;
  r : TRect;
  ndx : Integer;

begin
  n := FTestQuadTree.NextNode(FTestQuadTree.RootNode);
  while Assigned(n) do
  begin
    r.Left := Round((n.NodeArea.Left*2) + 360);
    r.Right := Round((n.NodeArea.Right*2) + 360);
    r.Top := Round(-(n.NodeArea.Top*2) + 180);
    r.Bottom := Round(-(n.NodeArea.Bottom*2) + 180);
    ndx := (n.NodeLevel mod 4);
    if n = n.ParentNode.ChildNodeNE then
      ndx := ndx + 0
    else if n = n.ParentNode.ChildNodeSE then
      ndx := ndx + 4
    else if n = n.ParentNode.ChildNodeSW then
      ndx := ndx + 8
    else if n = n.ParentNode.ChildNodeNW then
      ndx := ndx + 12;
    ACanvas.Pen.Color := ColorSet[ndx];
    ACanvas.Brush.Style := bsClear;
    ACanvas.Rectangle(r);
    n := FTestQuadTree.NextNode(n);
  end;
end;
function TForm1.PaintMouseHitTest(const X,Y : Integer; const AShift: TShiftState): TMouseRectHit;
const
  BorderInc = 1.0;
var
  llat, llon: Double;
  llr : TLazQuadNodeArea;
begin
  llat := -(Y / 2)+90;
  llon := (X /2)-180;
  Result := mrhNone;
  if ssAlt in AShift then
  begin
    if CoordIsInArea(llon, llat, FBaseRect,
        qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
    begin
      // Base Rectangle is Hit
      Result := mrhBaseArea;
      // Northwest edge
      llr.Top := FBaseRect.Top;
      llr.Bottom := FBaseRect.Top;
      llr.Left := FBaseRect.Left;
      llr.Right := FBaseRect.Right;
      IncLatLonRect(llr,BorderInc);
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhBaseNW;
        Exit;
      end;
      // NorthEast edge
      llr.Top := FBaseRect.Top;
      llr.Bottom := FBaseRect.Top;
      llr.Left := FBaseRect.Right;
      llr.Right := FBaseRect.Right;
      IncLatLonRect(llr,BorderInc);
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhBaseNE;
        Exit;
      end;
      // SouthEast edge
      llr.Top := FBaseRect.Bottom;
      llr.Left := FBaseRect.Right;
      llr.Bottom := FBaseRect.Bottom;
      llr.Right := FBaseRect.Right;
      IncLatLonRect(llr,BorderInc);
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhBaseSE;
        Exit;
      end;
      // SouthWest edge
      llr.Top := FBaseRect.Bottom;
      llr.Bottom := FBaseRect.Bottom;
      llr.Left := FBaseRect.Left;
      llr.Right := FBaseRect.Left;
      IncLatLonRect(llr,BorderInc);
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhBaseSW;
        Exit;
      end;
      llr.Left:=FBaseRect.Left;
      llr.Top:=FBaseRect.Top+BorderInc;
      llr.Right:=FBaseRect.Right;
      llr.Bottom:=FBaseRect.Top-BorderInc;
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhBaseTop;
        Exit;
      end;
      llr.Top:=FBaseRect.Bottom+BorderInc;
      llr.Bottom:=FBaseRect.Bottom-BorderInc;
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhBaseBottom;
        Exit;
      end;
      llr.Left:=FBaseRect.Left-BorderInc;
      llr.Top:=FBaseRect.Top;
      llr.Right:=FBaseRect.Left+BorderInc;
      llr.Bottom:=FBaseRect.Bottom;
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhBaseLeft;
        Exit;
      end;
      llr.Left:=FBaseRect.Right-BorderInc;
      llr.Right:=FBaseRect.Right+BorderInc;
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhBaseRight;
        Exit;
      end;
    end;
  end
  else
  begin
    if CoordIsInArea(llon, llat, FTestRect,
        qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
    begin
      // Test Rectangle is Hit
      Result := mrhTestArea;
      // Northwest edge
      llr.Top := FTestRect.Top;
      llr.Left := FTestRect.Left;
      llr.Bottom := FTestRect.Top;
      llr.Right := FTestRect.Left;
      IncLatLonRect(llr,BorderInc);
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhTestNW;
        Exit;
      end;
      // NorthEast edge
      llr.Top := FTestRect.Top;
      llr.Bottom := FTestRect.Top;
      llr.Left := FTestRect.Right;
      llr.Right := FTestRect.Right;
      IncLatLonRect(llr,BorderInc);
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhTestNE;
        Exit;
      end;
      // SouthEast edge
      llr.Top := FTestRect.Bottom;
      llr.Bottom := FTestRect.Bottom;
      llr.Left := FTestRect.Right;
      llr.Right := FTestRect.Right;
      IncLatLonRect(llr,BorderInc);
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhTestSE;
        Exit;
      end;
      // SouthWest edge
      llr.Top := FTestRect.Bottom;
      llr.Bottom := FTestRect.Bottom;
      llr.Left := FTestRect.Left;
      llr.Right := FTestRect.Left;
      IncLatLonRect(llr,BorderInc);
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhTestSW;
        Exit;
      end;
      llr.Left:=FTestRect.Left;
      llr.Top:=FTestRect.Top+BorderInc;
      llr.Right:=FTestRect.Right;
      llr.Bottom:=FTestRect.Top-BorderInc;
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhTestTop;
        Exit;
      end;
      llr.Top:=FTestRect.Bottom+BorderInc;
      llr.Bottom:=FTestRect.Bottom-BorderInc;
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhTestBottom;
        Exit;
      end;
      llr.Left:=FTestRect.Left-BorderInc;
      llr.Top:=FTestRect.Top;
      llr.Right:=FTestRect.Left+BorderInc;
      llr.Bottom:=FTestRect.Bottom;
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhTestLeft;
        Exit;
      end;
      llr.Left:=FTestRect.Right-BorderInc;
      llr.Right:=FTestRect.Right+BorderInc;
      if CoordIsInArea(llon, llat, llr,
          qydUpPlus, True, FTestQuadTree.RootNode.NodeArea) then
      begin
        Result := mrhTestRight;
        Exit;
      end;
    end;
  end;
end;
procedure TForm1.LatLonRectToGUI;
begin
  if FRectGUIUpdateInProgress then Exit;
  FRectGUIUpdateInProgress := True;
  try
{
    edtBaseNorthWestLat.Text := Format('%1.1f',[FBaseRect.NorthWestCorner.Lat]);
    edtBaseNorthWestLon.Text := Format('%1.1f',[FBaseRect.NorthWestCorner.Lon]);
    edtBaseSouthEastLat.Text := Format('%1.1f',[FBaseRect.SouthEastCorner.Lat]);
    edtBaseSouthEastLon.Text := Format('%1.1f',[FBaseRect.SouthEastCorner.Lon]);
    edtTestNorthWestLat.Text := Format('%1.1f',[FTestRect.NorthWestCorner.Lat]);
    edtTestNorthWestLon.Text := Format('%1.1f',[FTestRect.NorthWestCorner.Lon]);
    edtTestSouthEastLat.Text := Format('%1.1f',[FTestRect.SouthEastCorner.Lat]);
    edtTestSouthEastLon.Text := Format('%1.1f',[FTestRect.SouthEastCorner.Lon]);
}
  finally
    FRectGUIUpdateInProgress := False;
  end;
end;

end.

