{
uLoadCSV.pas

A helper Form to implement a user friendly loading of the time consuming load of
Data into the QuadTree

License: The same modified LGPL as the Free Pascal RTL
         See the file COPYING.modifiedLGPL for more details

Author: Ekkehard Domning (ekkehard@domning.eu)

}
unit uLoadCSV;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ComCtrls,
  ExtCtrls;

type
  { TLoadCSVLineIntoQuadTreeEvent a call back / event function to be called during loading }
  TLoadCSVLineIntoQuadTreeEvent = function(Sender : TObject; ALine : String) : Boolean of object;

  { TFormLoadCSV
    The form must be prepared before ShowModal and the results could be taken after the return.
    In between the OnLoadCSVLineIntoQuadTree is called for every line that should be imported into the
    QuadTree.
    The CSV Text must be set prior showing the form via the LoadList property
  }

  TFormLoadCSV = class(TForm)
    btnClose: TButton;
    Label1: TLabel;
    lblProgress: TLabel;
    pbLoadCSVProgress: TProgressBar;
    Timer1: TTimer;
    procedure btnCloseClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    FCloseOnLoadingDone : Boolean;
    FClearListOnClose : Boolean;
    FLoadedItemCount : Integer;
    FLoadList : TStringList;
    FOnLoadCSVLineIntoQuadTree : TLoadCSVLineIntoQuadTreeEvent;
    procedure SetLoadList(Value : TStrings);
    function GetLoadList : TStrings;
    procedure InternalClose;
  public
    property CloseOnLoadingDone : Boolean read FCloseOnLoadingDone write FCloseOnLoadingDone;
    property ClearListOnClose : Boolean read FClearListOnClose write FClearListOnClose;
    property OnLoadCSVLineIntoQuadTree : TLoadCSVLineIntoQuadTreeEvent read FOnLoadCSVLineIntoQuadTree write FOnLoadCSVLineIntoQuadTree;
    property LoadList : TStrings read GetLoadList write SetLoadList;
    property LoadedItemCount : Integer read FLoadedItemCount;
  end;

var
  FormLoadCSV: TFormLoadCSV;

implementation

{$R *.lfm}

{ TFormLoadCSV }

procedure TFormLoadCSV.FormCreate(Sender: TObject);
begin
  FLoadList := TStringList.Create;
  FClearListOnClose := True;
end;

procedure TFormLoadCSV.btnCloseClick(Sender: TObject);
begin
  InternalClose;
end;

procedure TFormLoadCSV.FormDestroy(Sender: TObject);
begin
  FLoadList.Free;
end;

procedure TFormLoadCSV.FormShow(Sender: TObject);
begin
  btnClose.Caption := 'Cancel';
  FLoadedItemCount := 0;
  Timer1.Enabled := True;
end;

procedure TFormLoadCSV.Timer1Timer(Sender: TObject);
var
  t0, t1 : Int64;
  s : String;
  cnt : Integer;
  mx,ict : Integer;
begin
  t0 := GetTickCount64;
  while FLoadedItemCount < FLoadList.Count do
  begin
    s := FLoadList[FLoadedItemCount];
    if Assigned(FOnLoadCSVLineIntoQuadTree) then
      if not FOnLoadCSVLineIntoQuadTree(Self,s) then
      begin
        Timer1.Enabled := False;
        lblProgress.Caption := Format('Error in Line %d of %d!',[FLoadedItemCount,FLoadList.Count]);
        btnClose.Caption := 'Done';
        if FCloseOnLoadingDone then
          InternalClose;
      end;
    Inc(FLoadedItemCount);
    mx := pbLoadCSVProgress.Max;
    ict := FLoadList.Count;
    if ict > 0 then
      cnt := Round(FLoadedItemCount/ict*mx)
    else
      cnt := 0;
    if cnt > pbLoadCSVProgress.Max then
      cnt := pbLoadCSVProgress.Max;
    pbLoadCSVProgress.Position := cnt;
    lblProgress.Caption := Format('%d of %d loaded',[FLoadedItemCount,FLoadList.Count]);
    t1 := GetTickCount64;
    if t1-t0 >= 20 then Break;
  end;
  if FLoadedItemCount >= FLoadList.Count then
  begin
    Timer1.Enabled := False;
    btnClose.Caption := 'Done';
    if FCloseOnLoadingDone then
      InternalClose;
  end;
end;

procedure TFormLoadCSV.SetLoadList(Value: TStrings);
begin
  FLoadList.Assign(Value);
  FLoadedItemCount := 0;
end;

function TFormLoadCSV.GetLoadList: TStrings;
begin
  Result := FLoadList;
end;

procedure TFormLoadCSV.InternalClose;
begin
  Timer1.Enabled := False;
  if FClearListOnClose then
    FLoadList.Clear;
  ModalResult := mrCancel;
  Close;
end;

end.

