{
ufrmLineRectangleHit.pas

A test-bed for the overlapping of lines and one rectangle.

License: The same modified LGPL as the Free Pascal RTL
         See the file COPYING.modifiedLGPL for more details

Author: Ekkehard Domning (ekkehard@domning.eu)

}
unit ufrmLineRectangleHit;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, ExtCtrls, Graphics, ComCtrls, StdCtrls,
  uLazQuadTreeGeometry, uMouseRectShifter;

type

  { TFrameLineRectangleHit }

  TFrameLineRectangleHit = class(TFrame)
    edtAnalysis: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    pnlPaintBox: TPanel;
    pbLineRectangleHit: TPaintBox;
    Panel1: TPanel;
    tbSpiralCount: TTrackBar;
    tbSpiralTurns: TTrackBar;
    procedure pbLineRectangleHitMouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure pbLineRectangleHitMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure pbLineRectangleHitMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pnlPaintBoxChangeBounds(Sender: TObject);
    procedure pbLineRectangleHitPaint(Sender: TObject);
    procedure tbSpiralCountChange(Sender: TObject);
    procedure tbSpiralTurnsChange(Sender: TObject);
  private
    FTestRect : TLQTRect;
    FPtWorld, FPbWorld : TLQTRect;
    FRectShifter : TMouseRectShifter;
    FShifterXInverseMode : Boolean; // Two Flags to ease the user interface for the flat and cylindrical world
    FShifterYInverseMode : Boolean;
    procedure SetupRectShifter;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;override;
  end;

implementation

{$R *.lfm}

{ TFrameLineRectangleHit }

procedure TFrameLineRectangleHit.pbLineRectangleHitPaint(Sender: TObject);
var
  ptcnt : Integer;
  ptloops : Double;
  pts : array of TLQTPoint = Nil;
  i, j : Integer;
  w : Integer;
  x,y : Double;
  ix0,iy0,ix1,iy1 : Integer;
  ptxs : array of Integer = Nil;
  ptys : array of Integer = Nil;
  li, li0 : TLineRectInteraction;
  linecnttotal, linecntinside, linecntintersect, linecntoutside : Integer;
  s : String;
begin
  linecnttotal := 0;
  linecntinside := 0;
  linecntintersect := 0;
  linecntoutside := 0;

  ptcnt := tbSpiralCount.Position;
  ptloops := tbSpiralTurns.Position / 4.0;
  w := pbLineRectangleHit.Width;
  // Paint the Background
  pbLineRectangleHit.Canvas.Pen.Color := clDkGray;
  pbLineRectangleHit.Canvas.Brush.Color := clSilver;
  pbLineRectangleHit.Canvas.Rectangle(0,0,w,w);

  // Paint the test rectangle
  TransformPoint(FTestRect.Left, FTestRect.Top,
                 FPtWorld, FPbWorld,
                 x,y);
  ix0 := Round(x);
  iy0 := Round(y);
  TransformPoint(FTestRect.Right, FTestRect.Bottom,
                 FPtWorld, FPbWorld,
                 x,y);
  ix1 := Round(x);
  iy1 := Round(y);
  pbLineRectangleHit.Canvas.Pen.Color := clDkGray;
  pbLineRectangleHit.Canvas.Brush.Color := clWhite;
  pbLineRectangleHit.Canvas.Rectangle(ix0,iy0,ix1,iy1);

  // Build the Spiral
  SetLength(pts,ptcnt);
  SetLength(ptxs,ptcnt);
  SetLength(ptys,ptcnt);
  for i := 0 to High(pts) do
  begin
    x := 0.0;
    y := (1 / ptcnt) * (i + 1);
    pts[i].Y := Sin(2.0 * PI * y * ptloops)*y;
    pts[i].X := Cos(2.0 * PI * y * ptloops)*y;
  end;
  for i := 0 to High(pts) do
  begin
    TransformPoint(pts[i].X, pts[i].Y,
                   FPtWorld, FPbWorld,
                   x,y);
    ix0 := Round(x);
    iy0 := Round(y);
    ptxs[i] := ix0;
    ptys[i] := iy0;
  end;
  // Now loop through all the lines, but paint the result grouped
  // so that the positive interactions are in front.
  for li := High(TLineRectInteraction) downto Low(TLineRectInteraction) do
  begin
    for i := 0 to High(pts) do
    begin
      for j := i-1 downto 0 do
      begin
        Inc(linecnttotal);
        li0 := LineRectInteraction(FTestRect,pts[i],pts[j]);
        if li0 <> li then Continue; // Not the Kind of hit we want, so continue
        case li0 of
          lriIntersecting :
            begin
              pbLineRectangleHit.Canvas.Pen.Color := clRed;
              Inc(linecntintersect);
            end;
          lriInside       :
            begin
              pbLineRectangleHit.Canvas.Pen.Color := clBlue;
              Inc(linecntinside);

            end
        else
          pbLineRectangleHit.Canvas.Pen.Color := clBlack;
          Inc(linecntoutside);

        end;
        pbLineRectangleHit.Canvas.Line(ptxs[i],ptys[i],ptxs[j],ptys[j]);
      end;
    end;
  end;
  s := Format('Lines: Total %d, Outside %d, Inside %d, Intersecting %d',
              [linecnttotal, linecntoutside, linecntinside, linecntintersect]);
  if edtAnalysis.Text <> s then
    edtAnalysis.Text := s;
end;

procedure TFrameLineRectangleHit.tbSpiralCountChange(Sender: TObject);
begin
  pbLineRectangleHit.Invalidate;
  pbLineRectangleHit.Repaint;
end;

procedure TFrameLineRectangleHit.tbSpiralTurnsChange(Sender: TObject);
begin
  pbLineRectangleHit.Invalidate;
  pbLineRectangleHit.Repaint;
end;

procedure TFrameLineRectangleHit.SetupRectShifter;
var
  x,y : Double;
  r : TRect;
begin
  TransformPoint(FTestRect.Left, FTestRect.Top,
                 FPtWorld, FPbWorld,
                 x,y);
  r.Left := Round(x);
  r.Top := Round(y);
  TransformPoint(FTestRect.Right, FTestRect.Bottom,
                 FPtWorld, FPbWorld,
                 x,y);
  r.Right := Round(x);
  r.Bottom := Round(y);
  FRectShifter.Clear;
  FRectShifter.AddRectangle(r,2,1);
end;

constructor TFrameLineRectangleHit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FTestRect.SetFrom(0.0,0.0,0.5,0.5);
  FPtWorld.SetFrom(-1.0,-1.0,1.0,1.0);
  FRectShifter := TMouseRectShifter.Create;
end;

destructor TFrameLineRectangleHit.Destroy;
begin
  FRectShifter.Free;
  inherited Destroy;
end;

procedure TFrameLineRectangleHit.pnlPaintBoxChangeBounds(Sender: TObject);
  var
  w, h : Integer;
  ratio : Double;
  pbh, pbw : Integer;
  pbc : Integer;
begin
  w := pnlPaintBox.Width;
  h := pnlPaintBox.Height;
  if (w < 2) or (h < 1) then
  begin
    w := 2;
    h := 1;
  end;
  ratio := w / h;
  if ratio > 1.0 then
  begin // wider than our background
    pbh := h;
    pbw := h;
    pbc := (w - h) div 2;
    pbLineRectangleHit.SetBounds(pbc,0,pbw,pbh);
  end
  else
  begin // taller than our background
    pbw := w;
    pbh := w;
    pbc := (h - w) div 2;
    pbLineRectangleHit.SetBounds(0,pbc,pbw,pbh);
  end;
  FPbWorld.SetFrom(0,0,pbw,pbh);
  SetupRectShifter;
end;

procedure TFrameLineRectangleHit.pbLineRectangleHitMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FRectShifter.OnMouseDown(Sender,Button,Shift,X,Y);
  // Reset the inverters
  FShifterXInverseMode := False;
  FShifterYInverseMode := False;
end;

procedure TFrameLineRectangleHit.pbLineRectangleHitMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
  // The following "PerformXXX" procedures, do the actial adjusting of the
  // side of the active rectangle.
  procedure PerformRightSide(const ANewX : Double; var ARect : TLQTRect);
  begin
    if FShifterXInverseMode then
    begin
      if (ANewX > ARect.Right) then
        FShifterXInverseMode := False;
    end
    else
    begin
      if (ANewX < ARect.Left) then
        FShifterXInverseMode := True;
    end;
    if FShifterXInverseMode then
      ARect.Left := ANewX
    else
      ARect.Right := ANewX;
  end;
  procedure PerformLeftSide(const ANewX : Double; var ARect : TLQTRect);
  begin
    if FShifterXInverseMode then
    begin
      if (ANewX < ARect.Left) then
        FShifterXInverseMode := False;
    end
    else
    begin
      if (ANewX > ARect.Right) then
        FShifterXInverseMode := True;
    end;
    if FShifterXInverseMode then
      ARect.Right := ANewX
    else
      ARect.Left := ANewX;
  end;
  procedure PerformTopSide(const ANewY : Double; var ARect : TLQTRect);
  begin
    if FShifterYInverseMode then
    begin
      if (ANewY < ARect.Top) then
        FShifterYInverseMode := False;
    end
    else
    begin
      if (ANewY > ARect.Bottom) then
        FShifterYInverseMode := True;
    end;
    if FShifterYInverseMode then
      ARect.Bottom := ANewY
    else
      ARect.Top := ANewY;
  end;
  procedure PerformBottomSide(const ANewY : Double; var ARect : TLQTRect);
  begin
    if FShifterYInverseMode then
    begin
      if (ANewY > ARect.Bottom) then
        FShifterYInverseMode := False;
    end
    else
    begin
      if (ANewY < ARect.Top) then
        FShifterYInverseMode := True;
    end;
    if FShifterYInverseMode then
      ARect.Top := ANewY
    else
      ARect.Bottom := ANewY;
  end;

var
  w, h : Double;
  lx,ly : Double;

  lRect0 : TLQTRect;
  lHitItem : TMouseHitItem;
begin
  w := pbLineRectangleHit.Width;
  FRectShifter.OnMouseMove(Sender,Shift,X,Y,lHitItem);
  if Assigned(lHitItem) then
  begin
    if pbLineRectangleHit.Cursor <> lHitItem.Cursor then
      pbLineRectangleHit.Cursor := lHitItem.Cursor;
    if lHitItem.MouseDownFlag then
    begin
      lRect0.Assign(FTestRect);
      // Transform it to the canvas world
      lRect0.Transform(FPtWorld, FPbWorld);
      // The HitItem is in CanvasWorld
      lx := (lHitItem.OrgX+lHitItem.MouseDeltaX);
      ly := (lHitItem.OrgY+lHitItem.MouseDeltaY);
      // Normate to world, just limit to the border
      NormatePointToWorld(lx,ly,FPbWorld,lqtwmFlat,lx,ly);

      case (lHitItem.Tag and $F) of
        tagRectArea :
          begin
            // Move the rectangle ...
            w := lRect0.Width;
            h := lRect0.Height;
            lRect0.Left := lx;
            lRect0.Top := ly;
            lRect0.Right := lRect0.Left + w;
            lRect0.Bottom := lRect0.Top + h;
            // ... but disallow shifting beyond the borders
            if lRect0.Right >= FPbWorld.Right then
              lRect0.Left := FPbWorld.Right - w;
            if lRect0.Bottom >= FPbWorld.Bottom then
              lRect0.Top := FPbWorld.Bottom - h;
          end;
        tagRectLeft :
          begin
            PerformLeftSide(lx,lRect0);
          end;
        tagRectTop :
          begin
            PerformTopSide(ly,lRect0);
          end;
        tagRectRight :
          begin
            PerformRightSide(lx,lRect0);
          end;
        tagRectBottom :
          begin
            PerformBottomSide(ly,lRect0);
          end;
        tagRectTopLeft :
          begin
            PerformTopSide(ly,lRect0);
            PerformLeftSide(lx,lRect0);
          end;
        tagRectTopRight :
          begin
            PerformTopSide(ly,lRect0);
            PerformRightSide(lx,lRect0);
          end;
        tagRectBottomRight :
          begin
            PerformBottomSide(ly,lRect0);
            PerformRightSide(lx,lRect0);
          end;
        tagRectBottomLeft :
          begin
            PerformBottomSide(ly,lRect0);
            PerformLeftSide(lx,lRect0);
          end;
      end;
      // Transform everything back
      lRect0.Transform(FPbWorld, FPtWorld);
      // and limit to the extent
      lRect0.NormateToWorld(FPtWorld);

      FTestRect.Assign(lRect0);
      pbLineRectangleHit.Invalidate;
      pbLineRectangleHit.Repaint;
    end;
  end
  else if pbLineRectangleHit.Cursor <> crDefault then
    pbLineRectangleHit.Cursor := crDefault;
end;

procedure TFrameLineRectangleHit.pbLineRectangleHitMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FRectShifter.OnMouseUp(Sender,Button,Shift,X,Y);
  SetupRectShifter;
end;


end.

