{
uLQTGeometryTestMain.pas

A test-bed for the uLazQuadTreeGeometry.pas tools, mostly rectangle related functions.

The three tab-sheets contains:
- A comprehensiv example of overlapping two rectangles.
- A threaded Test for the correct implementation of the function who calculates
  the overlapping (useful for veryfying the function after changes in the code).
- A simple line vs rectangle collision example.

License: The same modified LGPL as the Free Pascal RTL
         See the file COPYING.modifiedLGPL for more details

Author: Ekkehard Domning (ekkehard@domning.eu)

CAUTION: This form uses frames, to encapsulate parts of the program into different
         units. Sometimes, after changes in the GUI of the frames, the program starts
         with heavy exceptions. In this case find the frame in the object inspector and
         delete it. Then insert the frame again (second right icon in the "Standard"-tab
         of the components-bar). This should fix the problem.
         To ease this process, never rename inserted frames, instead use the name
         that is assigned by the IDE.
         Also, never assign any events in the inserted frame. Events should only assigned
         in the unit, or (rarely used) manually in the code.

}
unit uLQTGeometryTestMain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ComCtrls, ExtCtrls,
  StdCtrls, LCLType,
  LCLIntf, // For OpenURL
  uLazQuadTreeGeometry, uLQTRectOverlapThread,
  ufrmLQTRectOverlaptest, ufrmLineRectangleHit;

type

  { TForm1 }

  TForm1 = class(TForm)
    btnOverlapTestThread: TButton;
    FrameLineRectangleHit1: TFrameLineRectangleHit;
    FrameLQTRectOverlapTest1: TFrameLQTRectOverlapTest;
    gbOverlapTest: TGroupBox;
    imgNaturalEarth: TImage;
    imgEnigma: TImage;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    lblDomisLink: TLabel;
    lblEnigmaLink: TLabel;
    lblNaturalEarthLink: TLabel;
    lblStatusOverlapThread: TLabel;
    Label9: TLabel;
    PageControl1: TPageControl;
    Panel1: TPanel;
    Panel2: TPanel;
    pbOverlapTest: TProgressBar;
    tsAbout: TTabSheet;
    tsRectangleOverlap: TTabSheet;
    tsLineRectangleTest: TTabSheet;
    Timer1: TTimer;
    tsThreadedTest: TTabSheet;
    procedure btnOverlapTestThreadClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure lblLinkClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    FRectangleOverlapTestThread : TRectangleOverlapTestThread;
    procedure FinishRectangleOverlapTestThread;
  public

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }


procedure TForm1.btnOverlapTestThreadClick(Sender: TObject);
begin
  if Assigned(FRectangleOverlapTestThread) then
  begin
    if Application.MessageBox('Terminate RectangleOverlapTest?','Terminate Thread?',MB_YESNO) <> IDYes then
      Exit;
    FinishRectangleOverlapTestThread;
    btnOverlapTestThread.Caption := 'Start';
  end
  else
  begin
    FRectangleOverlapTestThread := TRectangleOverlapTestThread.Create(False);
    btnOverlapTestThread.Caption := 'Stop';
  end;
end;

procedure TForm1.FormActivate(Sender: TObject);
begin
  OnActivate := Nil;
  FrameLQTRectOverlapTest1.pnlPaintBoxChangeBounds(Nil);
  FrameLineRectangleHit1.pnlPaintBoxChangeBounds(Nil);
  imgNaturalEarth.Picture.Assign(FrameLQTRectOverlapTest1.CylindricBmp);
  imgEnigma.Picture.Assign(FrameLQTRectOverlapTest1.ToroidalBmp);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FinishRectangleOverlapTestThread;
end;

procedure TForm1.lblLinkClick(Sender: TObject);
begin
  if not (Sender is TLabel) then Exit;
  OpenURL(TLabel(Sender).Caption);
end;

procedure TForm1.Timer1Timer(Sender: TObject);
var
  d : Double;
begin
  if Assigned(FRectangleOverlapTestThread) then
  begin
    d := FRectangleOverlapTestThread.TestCounter / ExpectedTestCount;
    if d > 1.0 then
      d := 1.0;
    pbOverlapTest.Position := Round(d*1000.0);
    if not FRectangleOverlapTestThread.Terminated then
    begin
      lblStatusOverlapThread.Caption := 'Progress: '+IntToStr(FRectangleOverlapTestThread.TestCounter);
    end
    else
    begin
      if FRectangleOverlapTestThread.HitTestError then
        lblStatusOverlapThread.Caption := 'End with Error (Test: '+IntToStr(FRectangleOverlapTestThread.TestCounter)+')'
      else
        lblStatusOverlapThread.Caption := 'Success (Tests: '+IntToStr(FRectangleOverlapTestThread.TestCounter)+')';
      FinishRectangleOverlapTestThread;
      btnOverlapTestThread.Caption := 'Start';
    end;
  end;
end;

procedure TForm1.FinishRectangleOverlapTestThread;
begin
  if Assigned(FRectangleOverlapTestThread) then
  begin
    FRectangleOverlapTestThread.Terminate;
    if FRectangleOverlapTestThread.Suspended then
      FRectangleOverlapTestThread.Suspended := False;
    FRectangleOverlapTestThread.WaitFor;
    FreeAndNil(FRectangleOverlapTestThread);
  end;
end;

end.

